@@ -3,7 +3,7 @@
 # Cache implementaion with a Least Recently Used (LRU) replacement policy and a
 # basic dictionary interface.
 
-# Copyright (C) 2006, 2009, 2010  Jay Hutchinson
+# Copyright (C) 2006, 2009, 2010, 2011  Jay Hutchinson
 
 # This program is free software; you can redistribute it and/or modify it under
 # the terms of the GNU General Public License as published by the Free Software
@@ -42,8 +42,6 @@ class _dlnode(object):
         self.key = None
 
 
-
-
 class lrucache(object):
 
     def __init__(self, size, callback=None):
@@ -54,8 +52,8 @@ class lrucache(object):
 
 
         # Initalize the list with one empty node. Create a node and
-        # assign it to the 'head' variable, which represents the head node in the
-        # list.
+        # assign it to the 'head' variable, which represents the head node in
+        # the list.
         self.head = _dlnode()
         self.head.next = self.head
         self.head.prev = self.head
@@ -69,21 +67,18 @@ class lrucache(object):
     def __len__(self):
         return len(self.table)
 
-    # Does not call callback to write any changes!
     def clear(self):
-        self.table.clear()
-
-        node = self.head
-        for i in range(self.listSize):
+        for node in self.dli():
             node.key = None
             node.obj = None
-            node = node.next
+
+        self.table.clear()
 
 
     def __contains__(self, key):
         return key in self.table
-        # XXX Should this move the object to front of list? XXX
 
+    # Looks up a value in the cache without affecting cache order.
     def peek(self, key):
         # Look up the node
         node = self.table[key]
@@ -91,13 +86,12 @@ class lrucache(object):
 
 
     def __getitem__(self, key):
-
         # Look up the node
         node = self.table[key]
 
-        # Update the list ordering. Move this node so that is directly proceeds the
-        # head node. Then set the 'head' variable to it. This makes it the new head
-        # of the list.
+        # Update the list ordering. Move this node so that is directly proceeds
+        # the head node. Then set the 'head' variable to it. This makes it the
+        # new head of the list.
         self.mtf(node)
         self.head = node
 
@@ -106,9 +100,8 @@ class lrucache(object):
 
 
     def __setitem__(self, key, obj):
-
-        # First, see if any object is stored under 'key' in the cache already. If
-        # so we are going to replace that object with the new one.
+        # First, see if any object is stored under 'key' in the cache already.
+        # If so we are going to replace that object with the new one.
         if key in self.table:
 
             # Lookup the node
@@ -140,7 +133,7 @@ class lrucache(object):
         # If the node already contains something we need to remove the old key from
         # the dictionary.
         if node.key is not None:
-            if self.callback:
+            if self.callback is not None:
                 self.callback(node.key, node.obj)
             del self.table[node.key]
 
@@ -149,7 +142,7 @@ class lrucache(object):
         node.obj = obj
 
         # Add the node to the dictionary under the new key.
-        self.table[node.key] = node
+        self.table[key] = node
 
         # We need to move the node to the head of the list. The node is the tail
         # node, so it directly preceeds the head node due to the list being
@@ -241,7 +234,7 @@ class lrucache(object):
         for i in range(n):
             node = self.head.prev
             if node.key is not None:
-                if self.callback:
+                if self.callback is not None:
                     self.callback(node.key, node.obj)
                 del self.table[node.key]
 
@@ -259,12 +252,11 @@ class lrucache(object):
         self.listSize -= n
 
 
-    # This method adjusts the doubly linked list so that 'node' directly preeceds
-    # the 'head' node. Note that it works even if 'node' already directly
-    # preceeds the 'head' node or if 'node' is the 'head' node, in either case
-    # the order of the list is unchanged.
+    # This method adjusts the ordering of the doubly linked list so that 'node'
+    # directly precedes the 'head' node. Because of the order of operations, if
+    # 'node' already directly precedes the 'head' node or if 'node' is the
+    # 'head' node the order of the list will be unchanged.
     def mtf(self, node):
-
         node.prev.next = node.next
         node.next.prev = node.prev
 
@@ -274,105 +266,209 @@ class lrucache(object):
         node.next.prev = node
         node.prev.next = node
 
+    # This method returns an iterator that iterates over the non-empty nodes in
+    # the doubly linked list in order from the most recently to the least
+    # recently used.
     def dli(self):
         node = self.head
-        for i in range(len(self)):
+        for i in range(len(self.table)):
             yield node
             node = node.next
 
 
+         
 
-class lruwrap(object):
-    def __init__(self, store, size, writeback=False):
+class WriteThroughCacheManager(object):
+    def __init__(self, store, size):
         self.store = store
-        self.writeback = writeback
-
-        if not self.writeback:
-            self.cache = lrucache(size)
-        else:
-            self.dirty = set()
-            def callback(key, value):
-                if key in self.dirty:
-                    self.store[key] = value
-                    self.dirty.remove(key)
-            self.cache = lrucache(size, callback)
+        self.cache = lrucache(size)
 
     def __len__(self):
-        # XXX Need a way to efficiently return len() when writeback is turned
-        # on. If you really need the length you can call sync() then call
-        # len(self.store), but syncing all of the time kind of defeats the
-        # purpose of a writeback cache.
-        assert self.writeback == False
         return len(self.store)
 
+    # Returns/sets the size of the managed cache.
     def size(self, size=None):
-        self.cache.size(size)
+        return self.cache.size(size)
 
     def clear(self):
         self.cache.clear()
         self.store.clear()
-        if self.writeback:
-            self.dirty.clear()
 
     def __contains__(self, key):
-        # XXX Should this bring the key/value into the cache?
+        # Check the cache first. If it is there we can return quickly.
         if key in self.cache:
             return True
+
+        # Not in the cache. Might be in the underlying store.
         if key in self.store:
             return True
 
         return False
 
     def __getitem__(self, key):
+        # First we try the cache. If successful we just return the value. If not
+        # we catch KeyError and ignore it since that just means the key was not
+        # in the cache.
         try:
             return self.cache[key]
         except KeyError:
             pass
 
+        # It wasn't in the cache. Look it up in the store, add the entry to the
+        # cache, and return the value.
         value = self.store[key]
         self.cache[key] = value
         return value
 
     def __setitem__(self, key, value):
+        # Add the key/value pair to the cache and store.
         self.cache[key] = value
-
-        if self.writeback:
-            self.dirty.add(key)
-        else:
-            self.store[key] = value
+        self.store[key] = value
 
     def __delitem__(self, key):
-        if self.writeback:
-            found = False
-            try:
-                del self.cache[key]
-                found = True
+        # Write-through behavior cache and store should be consistent.
+        # Delete it from the store.
+        del self.store[key]
+        try:
+            # Ok, delete from the store was successful. It might also be in
+            # the cache, try and delete it. If not we catch the KeyError and
+            # ignore it.
+            del self.cache[key]
+        except KeyError:
+            pass
+
+    def __iter__(self):
+        return self.keys()
+
+    def keys(self):
+        return self.store.keys()
+
+    def values(self):
+        return self.store.values()
+
+    def items(self):
+        return self.store.items()
+        
+        
+
+class WriteBackCacheManager(object):
+    def __init__(self, store, size):
+        self.store = store
+        
+        # Create a set to hold the dirty keys.
+        self.dirty = set()
+
+        # Define a callback function to be called by the cache when a
+        # key/value pair is about to be ejected. This callback will check to
+        # see if the key is in the dirty set. If so, then it will update
+        # the store object and remove the key from the dirty set.
+        def callback(key, value):
+            if key in self.dirty:
+                self.store[key] = value
                 self.dirty.remove(key)
-            except KeyError:
-                pass
+                
+        # Create a cache and give it the callback function.
+        self.cache = lrucache(size, callback)
 
-            try:
-                del self.store[key]
-                found = True
-            except KeyError:
-                pass
+    # Returns/sets the size of the managed cache.
+    def size(self, size=None):
+        return self.cache.size(size)
 
-            if not found:  # If not found in cache or store, raise error.
-                raise KeyError
+    def clear(self):
+        self.cache.clear()
+        self.dirty.clear()
+        self.store.clear()
+        
+    def __contains__(self, key):
+        # Check the cache first, since if it is there we can return quickly.
+        if key in self.cache:
+            return True
 
-        else:  # Write-through behavior, cache and store should be consistent
+        # Not in the cache. Might be in the underlying store.
+        if key in self.store:
+            return True
+
+        return False
+
+    def __getitem__(self, key):
+        # First we try the cache. If successful we just return the value. If not
+        # we catch KeyError and ignore it since that just means the key was not
+        # in the cache.
+        try:
+            return self.cache[key]
+        except KeyError:
+            pass
+
+        # It wasn't in the cache. Look it up in the store, add the entry to the
+        # cache, and return the value.
+        value = self.store[key]
+        self.cache[key] = value
+        return value
+
+    def __setitem__(self, key, value):
+        # Add the key/value pair to the cache.
+        self.cache[key] = value
+        self.dirty.add(key)
+        
+    def __delitem__(self, key):
+        
+        found = False
+        try:
+            del self.cache[key]
+            found = True
+            self.dirty.remove(key)
+        except KeyError:
+            pass
+
+        try:
             del self.store[key]
-            try:
-                del self.cache[key]
-            except KeyError:
-                pass
+            found = True
+        except KeyError:
+            pass
+
+        if not found:  # If not found in cache or store, raise error.
+            raise KeyError
+
+        
+    def __iter__(self):
+        return self.keys()
+
+    def keys(self):
+        for key in self.store.keys():
+            if key not in self.dirty:
+                yield key
+            
+        for key in self.dirty:
+            yield key
+
+
+    def values(self):
+        for key, value in self.items():
+            yield value
+
+
+    def items(self):
+        for key, value in self.store.items():
+            if key not in self.dirty:
+                yield (key, value)
+            
+        for key in self.dirty:
+            value = self.cache.peek(key)
+            yield (key, value)
+
+
 
     def sync(self):
-        if self.writeback:
-            for key in self.dirty:
-                value = self.cache.peek(key)  # Doesn't change the cache's order
-                self.store[key] = value
-            self.dirty.clear()
+        # For each dirty key, peek at its value in the cache and update the
+        # store. Doesn't change the cache's order.
+        for key in self.dirty:
+            self.store[key] = self.cache.peek(key)
+        # There are no dirty keys now.
+        self.dirty.clear()
+
+    def flush(self):
+        self.sync()
+        self.cache.clear()
 
     def __enter__(self):
         return self
@@ -382,6 +478,18 @@ class lruwrap(object):
         return False
 
 
+
+
+
+def lruwrap(store, size, writeback=False):
+    if writeback:
+        return WriteBackCacheManager(store, size)
+    else:
+        return WriteThroughCacheManager(store, size)
+         
+         
+         
+
 class lrudecorator(object):
     def __init__(self, size):
         self.cache = lrucache(size)
